%%% The contents of this file are subject to the Erlang Public License,
%%% Version 1.0, (the "License"); you may not use this file except in
%%% compliance with the License. You may obtain a copy of the License at
%%% http://www.erlang.org/license/EPL1_0.txt
%%%
%%% Software distributed under the License is distributed on an "AS IS"
%%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%%% the License for the specific language governing rights and limitations
%%% under the License.
%%%
%%% The Original Code is xmerl-0.15
%%%
%%% The Initial Developer of the Original Code is Ericsson Telecom
%%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%%% Telecom AB. All Rights Reserved.
%%%
%%% Contributor(s):
%%%    Mickael Remond <mickael.remond@IDEALX.com>:
%%%    Johan Blom <johan.blom@mobilearts.se>
%%%    Richard Carlsson
%%%    Fredrik Linder
%%%
%%%----------------------------------------------------------------------
%%% #0.    BASIC INFORMATION
%%%----------------------------------------------------------------------
%%% File:       xmerl_scan.erl
%%% Author       : Ulf Wiger <ulf.wiger@ericsson.com>
%%% Description  : Simgle-pass XML scanner. See xmerl.hrl for data defs.
%%% 
%%% Modules used : ets, file, filename, io, lists, ucs, uri
%%% 
%%%----------------------------------------------------------------------
%% @doc 
%%     The XML parser is activated through 
%%     <tt>xmerl_scan:string/[1,2]</tt> or 
%%     <tt>xmerl_scan:file/[1,2]</tt>.
%%     It returns records of the type defined in xmerl.hrl.
%% See also <a href="xmerl_examples.html">tutorial</a> on customization
%% functions.
%% @type global_state(). <p>
%% The global state of the scanner, represented by the #xmerl_scanner{} record.
%% </p>
%% @type option_list(). <p>Options allows to customize the behaviour of the
%%     scanner.
%% See also <a href="xmerl_examples.html">tutorial</a> on customization
%% functions.
%% </p>
%% Possible options are:
%% <dl>
%%  <dt><code>{acc_fun, Fun}</code></dt>
%%    <dd>Call back function to accumulate contents of entity.</dd>
%%  <dt><code>{continuation_fun, Fun} |
%%            {continuation_fun, Fun, ContinuationState}</code></dt>
%%    <dd>Call back function to decide what to do if the scanner runs into eof
%%     before the document is complete.</dd>
%%  <dt><code>{event_fun, Fun} |
%%            {event_fun, Fun, EventState}</code></dt>
%%    <dd>Call back function to handle scanner events.</dd>
%%  <dt><code>{fetch_fun, Fun} |
%%            {fetch_fun, Fun, FetchState}</code></dt>
%%    <dd>Call back function to fetch an external resource.</dd>
%%  <dt><code>{hook_fun, Fun} |
%%            {hook_fun, Fun, HookState}</code></dt>
%%    <dd>Call back function to process the document entities once
%%     identified.</dd>
%%  <dt><code>{close_fun, Fun}</code></dt>
%%    <dd>Called when document has been completely parsed.</dd>
%%  <dt><code>{rules, ReadFun, WriteFun, RulesState} |
%%            {rules, Rules}</code></dt>
%%    <dd>Handles storing of scanner information when parsing.</dd>
%%  <dt><code>{user_state, UserState}</code></dt>
%%    <dd>Global state variable accessible from all customization functions</dd>
%%
%%  <dt><code>{fetch_path, PathList}</code></dt>
%%    <dd>PathList is a list of
%%     directories to search when fetching files. If the file in question
%%     is not in the fetch_path, the URI will be used as a file
%%     name.</dd>
%%  <dt><code>{prolog, Flag}</code></dt>
%%    <dd>'continue' (default) if scanning should continue after prolog in XML
%%    document, 'stop' otherwise.</dd>
%%  <dt><code>{space, Flag}</code></dt>
%%    <dd>'preserve' (default) to preserve spaces, 'normalize' to
%%    accumulate consecutive whitespace and replace it with one space.</dd>
%%  <dt><code>{line, Line}</code></dt>
%%    <dd>To specify starting line for scanning in document which contains
%%    fragments of XML.</dd>
%%  <dt><code>{namespace_conformant, Flag}</code></dt>
%%    <dd>Controls whether to behave as a namespace conformant XML parser,
%%    'false' (default) to not otherwise 'true'.</dd>
%%  <dt><code>{validation, Flag}</code></dt>
%%    <dd>Controls whether to process as a validating XML parser,
%%    'false' (default) to not otherwise 'true'.</dd>
%%  <dt><code>{quiet, Flag}</code></dt>
%%    <dd>Set to 'true' if xmerl should behave quietly and not output any info
%%    to standard output, 'false' (default) otherwise 'true'.</dd>
%%  <dt><code>{doctype_DTD, DTD}</code></dt>
%%    <dd>Allows to specify DTD name when it isn't available in the XML
%%    document.</dd>
%% </dl>


-module(xmerl_scan).
-vsn('0.19').
-date('03-09-16').


%% main API
-export([string/1, string/2,
	 file/1, file/2]).

%% access functions for various states
-export([user_state/1, user_state/2,
	 event_state/1, event_state/2,
	 hook_state/1, hook_state/2,
	 rules_state/1, rules_state/2,
	 fetch_state/1, fetch_state/2,
	 cont_state/1, cont_state/2]).

%% helper functions. To xmerl_lib ??
-export([accumulate_whitespace/4]).

%-define(debug, 1).
-include("xmerl.hrl").		% record def, macros
-include_lib("kernel/include/file.hrl").


-define(fatal(Reason, S),
	if
	    S#xmerl_scanner.quiet ->
		ok;
	    true ->
		ok=io:format("~p- fatal: ~p~n", [?LINE, Reason])
	end,
	fatal(Reason, S)).

-define(ustate(U, S), S#xmerl_scanner{user_state = U}).


%% Functions to access the various states

%%% @spec user_state(S::global_state()) -> global_state()
%%% @equiv user_state(UserState,S)
user_state(#xmerl_scanner{user_state = S}) -> S.

%%% @spec event_state(S::global_state()) -> global_state()
%%% @equiv event_state(EventState,S)
event_state(#xmerl_scanner{fun_states = #xmerl_fun_states{event = S}}) -> S.

%%% @spec hook_state(S::global_state()) -> global_state()
%%% @equiv hook_state(HookState,S)
hook_state(#xmerl_scanner{fun_states = #xmerl_fun_states{hook = S}}) -> S.

%%% @spec rules_state(S::global_state()) -> global_state()
%%% @equiv rules_state(RulesState,S)
rules_state(#xmerl_scanner{fun_states = #xmerl_fun_states{rules = S}}) -> S.

%%% @spec fetch_state(S::global_state()) -> global_state()
%%% @equiv fetch_state(FetchState,S)
fetch_state(#xmerl_scanner{fun_states = #xmerl_fun_states{fetch = S}}) -> S.

%%% @spec cont_state(S::global_state()) -> global_state()
%%% @equiv cont_state(ContinuationState,S)
cont_state(#xmerl_scanner{fun_states = #xmerl_fun_states{cont = S}}) -> S.


%%%% Functions to modify the various states

%%% @spec user_state(UserState, S::global_state()) -> global_state()
%%% @doc For controlling the UserState, to be used in a user function.
%%% See <a href="xmerl_examples.html">tutorial</a> on customization functions.
user_state(X, S) ->
    S#xmerl_scanner{user_state = X}.

%%% @spec event_state(EventState, S::global_state()) -> global_state()
%%% @doc For controlling the EventState, to be used in an event
%%% function, and called at the beginning and at the end of a parsed entity.
%%% See <a href="xmerl_examples.html">tutorial</a> on customization functions.
event_state(X, S = #xmerl_scanner{fun_states = FS}) ->
    FS1 = FS#xmerl_fun_states{event = X},
    S#xmerl_scanner{fun_states = FS1}.

%%% @spec hook_state(HookState, S::global_state()) -> global_state()
%%% @doc For controlling the HookState, to be used in a hook
%%% function, and called when the parser has parsed a complete entity.
%%% See <a href="xmerl_examples.html">tutorial</a> on customization functions.
hook_state(X, S = #xmerl_scanner{fun_states = FS}) ->
    FS1 = FS#xmerl_fun_states{hook = X},
    S#xmerl_scanner{fun_states = FS1}.

%%% @spec rules_state(RulesState, S::global_state()) -> global_state()
%%% @doc For controlling the RulesState, to be used in a rules
%%% function, and called when the parser store scanner information in a rules
%%% database.
%%% See <a href="xmerl_examples.html">tutorial</a> on customization functions.
rules_state(X, S = #xmerl_scanner{fun_states = FS}) ->
    FS1 = FS#xmerl_fun_states{rules = X},
    S#xmerl_scanner{fun_states = FS1}.

%%% @spec fetch_state(FetchState, S::global_state()) -> global_state()
%%% @doc For controlling the FetchState, to be used in a fetch
%%% function, and called when the parser fetch an external resource (eg. a DTD).
%%% See <a href="xmerl_examples.html">tutorial</a> on customization functions.
fetch_state(X, S = #xmerl_scanner{fun_states = FS}) ->
    FS1 = FS#xmerl_fun_states{fetch = X},
    S#xmerl_scanner{fun_states = FS1}.

%%% @spec cont_state(ContinuationState, S::global_state()) -> global_state()
%%% @doc For controlling the ContinuationState, to be used in a continuation
%%% function, and called when the parser encounters the end of the byte stream.
%%% See <a href="xmerl_examples.html">tutorial</a> on customization functions.
cont_state(X, S = #xmerl_scanner{fun_states = FS}) ->
    FS1 = FS#xmerl_fun_states{cont = X},
    S#xmerl_scanner{fun_states = FS1}.


%% @spec file(Filename::string()) -> xmlElement()
%% @equiv file(Name, [])
file(F) ->
    file(F, []).

%% @spec file(Filename::string(), Options::option_list()) -> xmlElement()
%%% @doc Parse file containing an XML document
file(F, Options) ->
    case int_file(F, Options) of
	{Res, Tail,S=#xmerl_scanner{close_fun=Close,
				    validation=ValidateResult}} ->
	    Close(S), % for side effects only - final state is dropped
	    {Res,Tail};
	{error, Reason} ->
	    {error, Reason};
	Other ->
	    {error, Other}
    end.

int_file(F, Options) ->
    case file:read_file(F) of
	{ok, Bin} ->
	    int_string(binary_to_list(Bin), Options, filename:dirname(F));
	Error ->
	    Error
    end.

%% @spec string(Text::list()) -> xmlElement()
%% @equiv string(Test, [])
string(Str) ->  
    string(Str, []).

%% @spec string(Text::list(),Options::option_list()) -> xmlElement()
%%% @doc Parse string containing an XML document
string(Str, Options) ->
    case int_string(Str, Options) of
	{Res, Tail, S = #xmerl_scanner{close_fun = Close,
				       validation=ValidateResult}} ->
	    Close(S),    % for side effects only - final state is dropped
	    {Res,Tail};
	{error, Reason} ->
	    {error, Reason};
	Other ->
	    {error, Other}
    end.

int_string(Str, Options) ->
    {ok, Cwd} = file:get_cwd(),
    S = initial_state0(Options,Cwd),
    scan_prolog(Str, S).

int_string(Str, Options, XMLBase) ->
    S = initial_state0(Options,XMLBase),
    scan_prolog(Str, S).
    


initial_state0(Options,XMLBase) ->
    initial_state(Options, #xmerl_scanner{
					  event_fun = fun event/2,
					  hook_fun = fun hook/2,
					  acc_fun = fun acc/3,
					  fetch_fun = fun fetch/2,
					  close_fun = fun close/1,
					  continuation_fun = fun cont/3,
					  rules_read_fun = fun rules_read/3,
					  rules_write_fun = fun rules_write/4,
					  directory = XMLBase
					 }).

initial_state([{event_fun, F}|T], S) ->
    initial_state(T, S#xmerl_scanner{event_fun = F});
initial_state([{event_fun, F, ES}|T], S) ->
    S1 = event_state(ES, S#xmerl_scanner{event_fun = F}),
    initial_state(T, S1);
initial_state([{acc_fun, F}|T], S) ->
    initial_state(T, S#xmerl_scanner{acc_fun = F});
initial_state([{hook_fun, F}|T], S) ->
    initial_state(T, S#xmerl_scanner{hook_fun = F});
initial_state([{hook_fun, F, HS}|T], S) ->
    S1 = hook_state(HS, S#xmerl_scanner{hook_fun = F}),
    initial_state(T, S1);
initial_state([{close_fun, F}|T], S) ->
    initial_state(T, S#xmerl_scanner{close_fun = F});
initial_state([{fetch_fun, F}|T], S) ->
    initial_state(T, S#xmerl_scanner{fetch_fun = F});
initial_state([{fetch_fun, F, FS}|T], S) ->
    S1 = fetch_state(FS, S#xmerl_scanner{fetch_fun = F}),
    initial_state(T, S1);
initial_state([{fetch_path, P}|T], S) ->
    initial_state(T, S#xmerl_scanner{fetch_path = P});
initial_state([{continuation_fun, F}|T], S) ->
    initial_state(T, S#xmerl_scanner{continuation_fun = F});
initial_state([{continuation_fun, F, CS}|T], S) ->
    S1 = cont_state(CS, S#xmerl_scanner{continuation_fun = F}),
    initial_state(T, S1);
initial_state([{rules, R}|T], S) ->
    initial_state(T, S#xmerl_scanner{rules = R,
				     keep_rules = true});
initial_state([{rules, Read, Write, RS}|T], S) ->
    S1 = rules_state(RS, S#xmerl_scanner{rules_read_fun = Read,
					 rules_write_fun = Write,
					 keep_rules = true}),
    initial_state(T, S1);
initial_state([{user_state, F}|T], S) ->
    initial_state(T, S#xmerl_scanner{user_state = F});
%initial_state([{directory, D}|T], S) ->
%    initial_state(T, S#xmerl_scanner{directory = D});
initial_state([{prolog, F}|T], S) ->
    initial_state(T, S#xmerl_scanner{prolog = F});
initial_state([{space, L}|T], S) ->
    initial_state(T, S#xmerl_scanner{space = L});
initial_state([{line, L}|T], S) ->
    initial_state(T, S#xmerl_scanner{line = L});
initial_state([{namespace_conformant, F}|T], S) when F==true; F==false ->
    initial_state(T, S#xmerl_scanner{namespace_conformant = F});
initial_state([{validation, F}|T], S) when F==true; F==false ->
    initial_state(T, S#xmerl_scanner{validation = F});
initial_state([{quiet, F}|T], S) when F==true; F==false ->
    initial_state(T, S#xmerl_scanner{quiet = F});
initial_state([{doctype_DTD,DTD}|T], S) ->
    initial_state(T,S#xmerl_scanner{doctype_DTD = DTD});
initial_state([], S = #xmerl_scanner{rules = undefined}) ->
    Tab = ets:new(rules, [set, public]),
    S#xmerl_scanner{rules = Tab};
initial_state([], S) ->
    S.


%%% -----------------------------------------------------
%%% Default modifier functions

%%% Hooks:
%%% - {element, Line, Name, Attrs, Content}
%%% - {processing_instruction, Line, Data}

hook(X, State) ->
    {X, State}.

%%% Events:
%%%
%%% #xmerl_event{event : started | ended,
%%%              line  : integer(),
%%%		 col   : integer(),
%%%              data}
%%%
%%% Data		Events
%%% document		started, ended
%%% #xmlElement		started, ended
%%% #xmlAttribute	ended
%%% #xmlPI		ended
%%% #xmlComment		ended
%%% #xmlText		ended
event(_X, S) ->
    S.

%% The acc/3 function can return either {Acc, S'} or {Acc', Pos', S'},
%% where Pos' can be derived from X#xmlElement.pos, X#xmlText.pos, or
%% X#xmlAttribute.pos (whichever is the current object type.)
%% The acc/3 function is not allowed to redefine the type of object
%% being defined, but _is_ allowed to either ignore it or split it 
%% into multiple objects (in which case {Acc',Pos',S'} should be returned.)
%% If {Acc',S'} is returned, Pos will be incremented by 1 by default.
%% Below is an example of an acceptable operation
acc(X = #xmlText{value = Text}, Acc, S) ->
    {[X#xmlText{value = lists:flatten(Text)}|Acc], S};
acc(X, Acc, S) ->
    {[X|Acc], S}.

fetch({system, URI}, S) ->
    fetch_URI(URI, S);
fetch({public, _PublicID, URI}, S) ->
    fetch_URI(URI, S).

%%% Always assume an external resource can be found locally! Thus don't bother
%%% fetching with e.g. HTTP. Returns the path where the resource is found.
%%% The path to the external resource is given by URI directly or the option 
%%% fetch_path (additional paths) or directory (base path to external resource)
fetch_URI(URI, S) ->
    %% assume URI is a filename
    Split = filename:split(URI),
    Filename = lists:last(Split),
    Fullname = 
	case Split of
	    ["file:",Name]->
		%% file:/dtd_name 
		filename:join(S#xmerl_scanner.directory, Name);
	    ["/"|Rest] when Rest /= [] ->
		%% absolute path name
		URI;
	    _ ->
		filename:join(S#xmerl_scanner.directory, URI)
	end,
    File = path_locate(S#xmerl_scanner.fetch_path, Filename, Fullname),
    ?dbg("fetch(~p) -> {file, ~p}.~n", [URI, File]),
    {ok, {file, File}, S}.

path_locate([Dir|Dirs], FN, FullName) ->
    F = filename:join(Dir, FN),
    case file:read_file_info(F) of
	{ok, #file_info{type = regular}} ->
	    F;
	_ ->
	    path_locate(Dirs, FN, FullName)
    end;
path_locate([], _FN, FullName) ->
    FullName.


cont(_F, Exception, US) ->
    Exception(US).

close(S) ->
    S.

%%%
%%% -----------------------------------------------------


%%% -----------------------------------------------------
%%% Scanner

%%%%%%% [22] Prolog
%%% prolog    ::=    XMLDecl? Misc* (doctypedecl Misc*)?
%%%
scan_prolog(Str, S = #xmerl_scanner{event_fun = Event,
				    line = L, col = C,
	                            validation=ValidateResult}) ->
    S1 = Event(#xmerl_event{event = started,
			    line = L,
			    col = C,
			    data = document}, S),

    {Res, Tail, S2} = scan_prolog(Str, S1, _StartPos = 1),
    
    S3 = #xmerl_scanner{} = Event(#xmerl_event{event = ended,
					       line = S2#xmerl_scanner.line,
					       col = S2#xmerl_scanner.col,
					       data = document}, S2),

    S4 = case ValidateResult of
	     false ->
		 cleanup(S3);
	     true ->
		 case xmerl_validate:validate(S3,Res) of
		     {error,Reason} ->
			 S3b=cleanup(S3),
			 ?fatal({failed_validation,Reason}, S3b);
		     _XML ->
			 cleanup(S3)
		 end
	 end,

    {Res, Tail, S4}.



scan_prolog([], S = #xmerl_scanner{continuation_fun = F}, Pos) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_prolog(MoreBytes, S1, Pos) end,
      fun(S1) -> {[], [], S1} end,
      S);
%%% More info on xml-stylesheet can be found at:
%%%   "Associating Style Sheets with XML documents", Version 1.0,
%%%   W3C Recommendation 29 June 1999 (http://www.w3.org/TR/xml-stylesheet/)
scan_prolog("<?xml-stylesheet" ++ T, S0=#xmerl_scanner{line=L,col=C},Pos) ->
    ?dbg("prolog(\"<?xml-stylesheet\")~n", []),
    ?bump_col(16),
    {_PI, T1, S1} = scan_pi(T, S, "xml-stylesheet",L,C+161,Pos,[]),
    scan_prolog(T1, S1, Pos+1);
scan_prolog("<?xml" ++ T, S0, Pos) ->
    ?dbg("prolog(\"<?xml\")~n", []),
    ?bump_col(5),
    {_, T1, S1} = scan_xml_decl(T, S),
    scan_prolog(T1, S1, Pos);
scan_prolog("<!--" ++ T, S, Pos) ->
    {_, T1, S1} = scan_comment(T, S, Pos, _Parents = [], _Lang = []),
    scan_prolog(T1, S1, Pos+1);
scan_prolog("<?" ++ T, S0, Pos) ->
    ?dbg("prolog(\"<?\")~n", []),
    ?bump_col(2),
    {_PI, T1, S1} = scan_pi(T, S, Pos),
    scan_prolog(T1, S1, Pos+1);
scan_prolog("<!DOCTYPE" ++ T, S0, Pos) ->
    ?dbg("prolog(\"<!DOCTYPE\")~n", []),
    ?bump_col(9),
    {T1, S1} = scan_doctype(T, S),
    scan_prolog(T1, S1, Pos);
scan_prolog(Str = "<!" ++ _, S, _Pos) ->
    ?dbg("prolog(\"<!\")~n", []),
    %% In e.g. a DTD, we jump directly to markup declarations
    scan_ext_subset(Str, S);
scan_prolog("<" ++ T, S0 = #xmerl_scanner{prolog=Mode,user_state=US}, Pos) ->
    ?dbg("prolog(\"<\")~n", []),
    case Mode of
	stop ->
	    {{US,"<" ++ T},[],S0};
	_ ->
	    %% OK, assume we're going directly for the data.
	    ?bump_col(1),
	    scan_element(T, S, Pos)
    end;
scan_prolog(T = [H|_], S, Pos) when ?whitespace(H) ->
    ?dbg("prolog(whitespace)~n", []),
    ?strip1,
    scan_prolog(T1, S1, Pos).


cleanup(S = #xmerl_scanner{keep_rules = false,
			   rules = Rules}) ->
    ets:delete(Rules),
    S#xmerl_scanner{rules = undefined};
cleanup(S) ->
    S.



%%%%%%% [23] XMLDecl
%%% XMLDecl     ::= '<?xml' VersionInfo EncodingDecl? SDDecl? S? '?>'
%%% VersionInfo ::= S 'version' Eq ("'" VersionNum "'" | '"' VersionNum '"')
scan_xml_decl(T, S) ->
    %% VersionInfo [24] is mandatory
    ?strip1,
    "version" ++ T2 = T1, S2=S1,
    {T3, S3} = scan_eq(T2, S2),
    {Vsn, T4, S4} = scan_xml_vsn(T3, S3),
    ?strip5,
    Attr = #xmlAttribute{name = version,
			 parents = [{xml, _XMLPos = 1}],
			 value = Vsn},
    scan_xml_decl(T5, S5, #xmlDecl{attributes = [Attr]}).

scan_xml_decl([], S = #xmerl_scanner{continuation_fun = F}, Decl) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_xml_decl(MoreBytes, S1, Decl) end,
      fun(S1) -> {[], [], S1} end,
      S);
scan_xml_decl("?>" ++ T, S0 = #xmerl_scanner{hook_fun = Hook,
					     event_fun = Event}, 
	      Decl0 = #xmlDecl{attributes = Attrs}) ->
    ?bump_col(2),
    ?strip1,
    Decl = Decl0#xmlDecl{attributes = lists:reverse(Attrs)},
    S2 = #xmerl_scanner{} = Event(#xmerl_event{event = ended,
					       line = S0#xmerl_scanner.line,
					       col = S0#xmerl_scanner.col,
					       data = Decl}, S1),
    {_Ret, S3} = Hook(Decl, S2),
    {[], T1, S3};
scan_xml_decl("encoding" ++ T, S0 = #xmerl_scanner{event_fun = Event},
	      Decl0 = #xmlDecl{attributes = Attrs}) ->
    %% [80] EncodingDecl
    ?bump_col(8),
    {T1, S1} = scan_eq(T, S),
    {EncName, T2, S2} = scan_enc_name(T1, S1),
    Attr = #xmlAttribute{name = encoding,
			 parents = [{xml, _XMLPos = 1}],
			 value = EncName},
    Decl = Decl0#xmlDecl{encoding = EncName,
			 attributes = [Attr|Attrs]},
    S3 = #xmerl_scanner{} = Event(#xmerl_event{event = ended, 
					       line = S0#xmerl_scanner.line, 
					       col = S0#xmerl_scanner.col,
					       data = Attr}, S2),
    scan_xml_decl(T2, S3, Decl);
scan_xml_decl("standalone" ++ T,S0 = #xmerl_scanner{event_fun = Event},
	      Decl0 = #xmlDecl{attributes = Attrs}) ->
    %%[32] SDDecl::=
    ?bump_col(10),
    {T1, S1} = scan_eq(T, S),
    {StValue,T2,S2}=scan_standalone_value(T1,S1),
    Attr = #xmlAttribute{name = standalone,
			 parents = [{xml, _XMLPos = 1}],
			 value = StValue},
    Decl = Decl0#xmlDecl{standalone = StValue,
			 attributes = [Attr|Attrs]},
    S3 = #xmerl_scanner{} = Event(#xmerl_event{event = ended, 
					       line = S0#xmerl_scanner.line, 
					       col = S0#xmerl_scanner.col,
					       data = Attr}, S2),
    scan_xml_decl(T2, S3, Decl).

scan_standalone_value("'yes'" ++T,S0)->
    ?bump_col(5),
    {'yes',T,S};
scan_standalone_value("'no'" ++T,S0) ->
    ?bump_col(4),
    {'no',T,S}.


%%%%%%% [81] EncName
scan_enc_name([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_enc_name(MoreBytes, S1) end,
      fun(S1) -> ?fatal(expected_encoding_name, S1) end,
      S);
scan_enc_name([H|T], S0) when H >= $"; H =< $' -> 
    ?bump_col(1),
    scan_enc_name(T, S, H, []).


scan_enc_name([], S = #xmerl_scanner{continuation_fun = F}, Delim, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_enc_name(MoreBytes, S1, Delim, Acc) end,
      fun(S1) -> ?fatal(expected_encoding_name, S1) end,
      S);
scan_enc_name([H|T], S0, H, Acc) ->
    ?bump_col(1),
    ?strip1,
    {lists:reverse(Acc), T1, S1};
scan_enc_name([H|T], S0, Delim, Acc) when H >= $a, H =< $z ->
    ?bump_col(1),
    scan_enc_name(T, S, Delim, [H|Acc]);
scan_enc_name([H|T], S0, Delim, Acc) when H >= $A, H =< $Z ->
    ?bump_col(1),
    scan_enc_name(T, S, Delim, [H|Acc]);
scan_enc_name([H|T], S0, Delim, Acc) when H >= $0, H =< $9 ->
    ?bump_col(1),
    scan_enc_name(T, S, Delim, [H|Acc]);
scan_enc_name([H|T], S0, Delim, Acc) when H == $.; H == $_; H == $- ->
    ?bump_col(1),
    scan_enc_name(T, S, Delim, [H|Acc]).


%%%%%%% [26] VersionNum
%%% VersionNum    ::=    ([a-zA-Z0-9_.:] | '-')+
scan_xml_vsn([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_xml_vsn(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_xml_vsn([H|T], S) when H==$"; H==$'->
    xml_vsn(T, S#xmerl_scanner{col = S#xmerl_scanner.col+1}, H, []).

xml_vsn([], S = #xmerl_scanner{continuation_fun = F}, Delim, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> xml_vsn(MoreBytes, S1, Delim, Acc) end, 
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
xml_vsn([H|T], S = #xmerl_scanner{col = C}, H, Acc) ->
    {lists:reverse(Acc), T, S#xmerl_scanner{col = C+1}};
xml_vsn([H|T], S = #xmerl_scanner{col = C}, 
	Delim, Acc) when H >= $a, H =< $z ->
    xml_vsn(T, S#xmerl_scanner{col = C+1}, Delim, [H|Acc]);
xml_vsn([H|T], S = #xmerl_scanner{col = C}, 
	Delim, Acc) when H >= $A, H =< $Z ->
    xml_vsn(T, S#xmerl_scanner{col = C+1}, Delim, [H|Acc]);
xml_vsn([H|T], S = #xmerl_scanner{col = C}, 
	Delim, Acc) when H >= $0, H =< $9 ->
    xml_vsn(T, S#xmerl_scanner{col = C+1}, Delim, [H|Acc]);
xml_vsn([H|T], S = #xmerl_scanner{col = C}, Delim, Acc) ->
    case lists:member(H, "_.:-") of
	true ->
	    xml_vsn(T, S#xmerl_scanner{col = C+1}, Delim, [H|Acc]);
	false ->
	    ?fatal({invalid_vsn_char, H}, S)
    end.

%%%%%%% [16] PI

scan_pi([], S = #xmerl_scanner{continuation_fun = F}, Pos) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_pi(MoreBytes, S1, Pos) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_pi(Str = [H1,H2,H3 | _T], 
	S = #xmerl_scanner{line = L, col = C}, Pos) when H1 == $x; H1 == $X ->
    %% names beginning with [xX][mM][lL] are reserved for future use.
    if H2 == $m; H2 == $M ->
	    if H3 == $l; H3 == $L ->
		    ?fatal({invalid_target_name, lists:sublist(Str, 1, 6)}, S);
	       true ->
		    ok
	    end;
       true ->
	    ok
    end,
    {Target, _NamespaceInfo, T1, S1} = scan_name_no_colons(Str, S),
    scan_pi(T1, S1, Target, L, C, Pos, []);
scan_pi(Str, S = #xmerl_scanner{line = L, col = C}, Pos) ->
    {Target, _NamespaceInfo, T1, S1} = scan_name(Str, S),
    scan_pi(T1, S1, Target, L, C, Pos, []).


scan_pi([], S = #xmerl_scanner{continuation_fun = F}, Target, 
	L, C, Pos, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_pi(MoreBytes, S1, Target, L, C, Pos, Acc) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_pi("?>" ++ T, S0 = #xmerl_scanner{hook_fun = Hook,
				       event_fun = Event}, 
	Target, L, C, Pos, Acc) ->
    ?bump_col(2),
    PI = #xmlPI{name = Target,
		pos = Pos,
		value = lists:reverse(Acc)},
    S1 = #xmerl_scanner{} = Event(#xmerl_event{event = ended,
					       line = L,
					       col = C,
					       data = PI}, S),
    {Ret, S2} = Hook(PI, S1),
    {Ret, T, S2};
scan_pi([H|T], S0, Target, L, C, Pos, Acc) ->
    ?bump_col(1),
    scan_pi(T, S, Target, L, C, Pos, [H|Acc]).



scan_doctype([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_doctype(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_doctype(T, S) ->
    ?strip1,
    {DTName, _NamespaceInfo, T2, S2} = scan_name(T1, S1),
    ?strip3,
    scan_doctype1(T3, S3#xmerl_scanner{doctype_name =  DTName}).


scan_doctype1([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_doctype1(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_doctype1("PUBLIC" ++ T, S0) ->
    ?bump_col(6),
    ?strip1,
    {PIDL, T2, S2} = scan_pubid_literal(T1, S1),
    ?strip3,
    {SL, T4, S4} = scan_system_literal(T3, S3),
    ?strip5,
    scan_doctype2(T5, S5, {public, PIDL, SL});
scan_doctype1("SYSTEM" ++ T, S0) ->
    ?bump_col(6),
    ?strip1,
    {SL, T2, S2} = scan_system_literal(T1, S1),
    ?strip3,
    scan_doctype2(T3, S3, {system, SL});
scan_doctype1(T, S) ->
    scan_doctype2(T, S, undefined).


scan_doctype2([], S = #xmerl_scanner{continuation_fun = F},DTD) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_doctype2(MoreBytes, S1, DTD) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_doctype2("[" ++ T, S0, DTD) ->
    ?bump_col(1),
    ?strip1,
    scan_doctype3(T1, S1, DTD);
scan_doctype2(">" ++ T, S0, DTD) ->
    ?bump_col(1),
    ?strip1,
    S2 = fetch_DTD(DTD, S1),
    {T1, S2}.

scan_doctype3([], S = #xmerl_scanner{continuation_fun = F},DTD) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_doctype3(MoreBytes, S1,DTD) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_doctype3("%" ++ T, S0, DTD) ->
    ?bump_col(1),
    {PERefName, T1, S1} = scan_pe_reference(T, S),
    case expand_pe_reference(PERefName, S1) of
	{system, _} = Name ->
	    S2 = fetch_DTD(Name, S1),
	    scan_doctype3(T1, S2, DTD);
	{public, _} = Name ->
	    S2 = fetch_DTD(Name, S1),
	    scan_doctype3(T1, S2, DTD);
	{public, _, _} = Name ->
	    S2 = fetch_DTD(Name, S1),
	    scan_doctype3(T1, S2, DTD);
	ExpRef when list(ExpRef) -> % Add spaces, see Section 4.4.8
	    scan_doctype3(" "++ExpRef++" "++T1,S1,DTD)
    end;
scan_doctype3("]" ++ T, S0, DTD) ->
    ?bump_col(1),
    ?strip1,
    S2 = fetch_DTD(DTD, S1),
    ">" ++ T2 = T1,
    {T2, S2};
scan_doctype3(T, S, DTD) ->
    {_, T1, S1} = scan_markup_decl(T, S),
    scan_doctype3(T1, S1, DTD).

fetch_DTD(undefined, #xmerl_scanner{doctype_DTD=URI}=S) when list(URI)->
    %% allow to specify DTD name when it isn't available in xml stream
    fetch_DTD({system,URI},S);
fetch_DTD(undefined, S) ->
    S;
fetch_DTD(DTDSpec, S)-> 
    case fetch_and_parse(DTDSpec,S,[]) of
	NewS when record(NewS,xmerl_scanner) ->
	    NewS;
	{_Res,_Tail,_Sx} -> % Continue with old scanner data, result in Rules
	    S
%	    #xmerl_scanner{} = (S#xmerl_scanner.close_fun)(S)
    end.

% fetch_DTD(DTDSpec, S=#xmerl_scanner{})-> 
%     Rules=case DTDSpec of
% 	      {_,URI0} ->   list_to_atom(URI0);
% 	      {_,_,URI0} -> list_to_atom(URI0)
% 	  end,
%     case ets:info(Rules,size) of
% 	undefined->
% 	    io:format("fetch_DTD ~p ~p~n",[Rules,S#xmerl_scanner.rules]),
% 	    ets:new(Rules,[set,public,named_table]),
% 	    fetch_DTD2(DTDSpec,S#xmerl_scanner{rules=Rules});
% 	_ ->
% 	    S#xmerl_scanner{rules=Rules}
%     end.


fetch_and_parse(ExtSpec,S=#xmerl_scanner{fetch_fun=Fetch,rules=Rules},
		Options0) ->
    case Fetch(ExtSpec, S) of
	{ok, NewS} -> 
	    %% For backward compatibility only. This will be removed later!!
	    NewS;
	{ok, not_fetched,NewS} ->
	    NewS;
	{ok, DataRet, NewS = #xmerl_scanner{user_state = UState,
					    event_fun = Event,
					    hook_fun = Hook,
					    fetch_fun = Fetch1,
					    close_fun = Close1,
					    continuation_fun = Cont,
					    acc_fun = Acc,
					    rules_read_fun = Read,
					    rules_write_fun = Write
					   }} ->
	    EvS = event_state(NewS),
	    HoS = hook_state(NewS),
	    FeS = fetch_state(NewS),
	    CoS = cont_state(NewS),
	    Options = Options0++[{user_state, UState},
				 {rules, Rules},
				 {event_fun, Event, EvS},
				 {hook_fun, Hook, HoS},
				 {fetch_fun, Fetch1, FeS},
				 {close_fun, Close1},
				 {continuation_fun, Cont, CoS},
				 {rules, Read, Write, ""},
				 {acc_fun, Acc}],
	    case DataRet of
		{file, F} ->
		    int_file(F, Options);
%		    {_Res,_Tail,Sx}=int_file(F, Options),
%		    #xmerl_scanner{} = Close1(Sx);
		{string, String} ->
		    int_string(String, Options)
%		    {_Res,_Tail,Sx}=int_string(String, Options),
%		    ?dbg("DTD string -> US = ~p~n", [String]),
%		    #xmerl_scanner{} = Close1(Sx)
	    end;
	Error ->
	    ?fatal({error_fetching_DTD, {ExtSpec, Error}}, S)
    end.


%%%%%%% [30] extSubSet

scan_ext_subset([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_ext_subset(MoreBytes, S1) end,
      fun(S1) -> {[], [], S1} end,
      S);
scan_ext_subset("%" ++ T, S0) ->
    ?bump_col(1),
    {PERefName, T1, S1} = scan_pe_reference(T, S0),
    S2 = expand_external_pe_reference(PERefName, S1),
    {_, T2, S3} = strip(T1,S2),
    scan_ext_subset(T2, S3);
scan_ext_subset("<![" ++ T, S0) ->
    ?bump_col(3),
    ?strip1,
    {_, _T2, _S2} = scan_conditional_sect(T1, S1);
scan_ext_subset(T, S) ->
    {_, T1, S1} = scan_markup_decl(T, S),
    scan_ext_subset(T1, S1).


%%%%%%% [61] ConditionalSect

scan_conditional_sect([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_conditional_sect(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_conditional_sect("IGNORE" ++ T, S0) ->
    ?bump_col(6),
    scan_ignore(T, S);
scan_conditional_sect("INCLUDE" ++ T, S0) ->
    ?bump_col(7),
    ?strip1,
    scan_include(T1, S1).


%%%%%%% [63] IgnoreSect

scan_ignore([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_ignore(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_ignore("<![" ++ T, S0) ->
    %% nested conditional section. Topmost condition is ignore, though
    ?bump_col(3),
    {_, T1, S1} = scan_ignore(T, S),
    scan_ignore(T1, S1);
scan_ignore("]]>" ++ T, S0) ->
    ?bump_col(3),
    {[], T, S}.


%%%%%%% [62] IncludeSect

scan_include([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_include(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_include("]]>" ++ T, S0) ->
    ?bump_col(3),
    {[], T, S};
scan_include("%" ++ T, S0) ->
    ?bump_col(1),
    {PERefName, T1, S1} = scan_pe_reference(T, S),
    ExpRef = expand_pe_reference(PERefName, S1),
    scan_include(ExpRef ++ T1, S1);
scan_include("<![" ++ T, S0) ->
    ?bump_col(3),
    ?strip1,
    {_, _T2, _S2} = scan_conditional_sect(T1, S1);
scan_include(T, S) ->
    {_, T1, S1} = scan_markup_decl(T, S),
    scan_include(T1, S1).


%%%%%%% [29] markupdecl
%%%%%%% [45] elementdecl

%% Validity constraint: Unique Type Declaration: No element type may be
%% declared more than once.
%%
scan_markup_decl([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_markup_decl(MoreBytes, S1) end,
      fun(S1) -> {[], [], S1} end,
      S);
scan_markup_decl("<!--" ++ T, S) ->
    {_, T1, S1} = scan_comment(T, S),
    ?strip2;
scan_markup_decl("<!ELEMENT" ++ T, 
		 #xmerl_scanner{rules_read_fun = Read,
				rules_write_fun = Write} = S0) ->
    ?bump_col(9),
    ?strip1,
    {Ename, _NamespaceInfo, T2, S2} = scan_name(T1, S1),
    case Read(elem_def, Ename, S2) of
	PrevDef when PrevDef /= undefined ->
	    ?fatal({already_defined, Ename}, S2);
	_ ->
	    ok
    end,
    ?strip3,
    {Edef, T4, S4} = scan_contentspec(T3, S3),
    ?strip5,
    ">" ++ T6 = T5,
    S6 = Write(elem_def, Ename, #xmlElement{name = Ename,
					    content = Edef}, S5),
    ?strip7;
scan_markup_decl("<!ENTITY" ++ T, S0) ->
    %% <!ENTITY [%] entity.name NDATA notation.name>
    %% <!ENTITY [%] entity.name "replacement text">
    %% <!ENTITY [%] entity.name SYSTEM "system.identifier">
    %% <!ENTITY [%] entity.name PUBLIC public.identifier "system.identifier">
    ?bump_col(8),
    ?strip1,
    {T2, S2} = scan_entity(T1, S1),
    ?strip3;
scan_markup_decl("<!NOTATION" ++ T, S0) ->
    %% <!NOTATION notation.name "public.identifier" "helper.application">
    ?bump_col(10),
    ?strip1,
    {T2, S2} = scan_notation_decl(T1, S1),
    ?strip3;
scan_markup_decl("<!ATTLIST" ++ T, 
		 #xmerl_scanner{rules_read_fun = Read,
				rules_write_fun = Write} = S0) ->
    %% <!ATTLIST Ename ( AttrName Type Value )*>
    ?bump_col(9),
    ?strip1,
    {Ename, _NamespaceInfo, T2, S2} = scan_name(T1, S1),
    case Read(elem_def, Ename, S2) of
	undefined ->
	    ?fatal({unknown_element, Ename}, S2);
	Edef = #xmlElement{attributes = OldAttrs} ->
	    ?strip3,
	    {Attributes, T4, S4} = scan_attdef(T3, S3),
	    NewAttrs = update_attributes(Attributes, OldAttrs),
	    NewEdef = Edef#xmlElement{attributes = NewAttrs},
	    S5 = Write(elem_def, Ename, NewEdef, S4),
	    T5 = T4,
	    ?strip6
    end.

update_attributes(NewAttrs, OldAttrs) ->
    update_attributes(NewAttrs, OldAttrs, lists:reverse(OldAttrs)).

update_attributes([A = {Name, _Type, _Default}|Attrs], OldAttrs, Acc) ->
    case lists:keymember(Name, 1, OldAttrs) of
	true ->
	    update_attributes(Attrs, lists:keydelete(Name, 1, OldAttrs),
			      Acc);
	false ->
	    update_attributes(Attrs, lists:keydelete(Name, 1, OldAttrs), 
			      [A|Acc])
    end;
update_attributes([], _, Acc) ->
    lists:reverse(Acc).


%%%%%%% [53] AttDef

scan_attdef([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_attdef(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_attdef(T, S) ->
    scan_attdef(T, S, _AttrAcc = []).


scan_attdef([], S = #xmerl_scanner{continuation_fun = F}, Attrs) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_attdef(MoreBytes, S1, Attrs) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_attdef(">" ++ T, S0, Attrs) ->
    ?bump_col(1),
    {lists:reverse(Attrs), T, S};
scan_attdef("%" ++ T, S0, Attrs) ->
    ?bump_col(1),
    {PERefName, T1, S1} = scan_pe_reference(T, S0),
    ExpRef = expand_pe_reference(PERefName, S1),
    scan_attdef(ExpRef ++ T1, S1, Attrs);
scan_attdef(T, S, Attrs) ->
    {AttName, _NamespaceInfo, T1, S1} = scan_name(T, S),
    ?strip2,
    {AttType, T3, S3} = scan_att_type(T2, S2),
    ?strip4,
    {DefaultDecl, T5, S5} = scan_default_decl(T4, S4, AttType),
    ?strip6,
    scan_attdef(T6, S6, [{AttName, AttType, DefaultDecl}|Attrs]).


%%% [54] StringType
scan_att_type([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_att_type(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_att_type("CDATA" ++ T, S0) ->
    ?bump_col(5),
    {'CDATA', T, S};
%%% [55] TokenizedType
scan_att_type("IDREFS" ++ T, S0) ->
    ?bump_col(6),
    {'IDREFS', T, S};
scan_att_type("IDREF" ++ T, S0) ->
    ?bump_col(5),
    {'IDREF', T, S};
scan_att_type("ID" ++ T, S0) ->
    ?bump_col(2),
    {'ID', T, S};
scan_att_type("ENTITY" ++ T, S0) ->
    ?bump_col(6),
    {'ENTITY', T, S};
scan_att_type("ENTITIES" ++ T, S0) ->
    ?bump_col(8),
    {'ENTITIES', T, S};
scan_att_type("NMTOKENS" ++ T, S0) ->
    ?bump_col(8),
    {'NMTOKENS', T, S};
scan_att_type("NMTOKEN" ++ T, S0) ->
    ?bump_col(7),
    {'NMTOKEN', T, S};
%%% [57] EnumeratedType
scan_att_type("NOTATION" ++ T, S0) ->
    ?bump_col(8),
    ?strip1,
    "(" ++ T2 = T1,
    S2 = S1,
    ?strip3,
    {Name, _NamespaceInfo, T4, S4} = scan_name(T3, S3),
    notation_exists(Name, S4),
    scan_notation_type(T4, S4, [Name]);
scan_att_type("(" ++ T, S0) ->
    ?bump_col(1),
    ?strip1,
    {NmToken, _NamespaceInfo, T2, S2} = scan_nmtoken(T1, S1),
    ?strip3,
    scan_enumeration(T3, S3, [NmToken]);
scan_att_type("%" ++ T, S0) ->
    ?bump_col(1),
    {PERefName, T1, S1} = scan_pe_reference(T, S0),
    ExpRef = expand_pe_reference(PERefName, S1),
    {ExpRef,T1,S1}.

%%% [58] NotationType

scan_notation_type([], S = #xmerl_scanner{continuation_fun = F}, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_notation_type(MoreBytes, S1, Acc) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_notation_type(")" ++ T, S0, Acc) ->
    ?bump_col(1),
    {{notation, lists:reverse(Acc)}, T, S};
scan_notation_type("|" ++ T, S0, Acc) ->
    ?bump_col(1),
    ?strip1,
    {Name, _NamespaceInfo, T2, S2} = scan_name(T1, S1),
    notation_exists(Name, S2),
    scan_notation_type(T2, S2, [Name | Acc]).

%%% Validity constraint for NotationType: 
%%% The notation names must have been declared.
notation_exists(Name, #xmerl_scanner{rules_read_fun = Read} = S) ->
    case Read(notation, Name, S) of
	undefined ->
	    ?fatal({unknown_notation, Name}, S);
	_Value ->
	    ok
    end.

%%% [59] Enumeration

scan_enumeration([], S = #xmerl_scanner{continuation_fun = F}, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_enumeration(MoreBytes, S1, Acc) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_enumeration(")" ++ T, S0, Acc) ->
    ?bump_col(1),
    {{enumeration, lists:reverse(Acc)}, T, S};
scan_enumeration("|" ++ T, S0, Acc) ->
    ?bump_col(1),
    ?strip1,
    {NmToken, _NamespaceInfo, T2, S2} = scan_nmtoken(T1, S1),
    ?strip3,
    scan_enumeration(T3, S3, [NmToken|Acc]).


%%%%%%% [60] DefaultDecl

scan_default_decl([], S = #xmerl_scanner{continuation_fun = F}, Type) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_default_decl(MoreBytes, S1, Type) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_default_decl("#REQUIRED" ++ T, S0, _Type) ->
    ?bump_col(9),
    {'#REQUIRED', T, S};
scan_default_decl("#IMPLIED" ++ T, S0, _Type) ->
    ?bump_col(8),
    {'#IMPLIED', T, S};
scan_default_decl("#FIXED" ++ T, S0, Type) ->
    ?bump_col(6),
    ?strip1,
    default_value(T1, S1, Type);
scan_default_decl(Str, S, Type) ->
    default_value(Str, S, Type).


%% There is room here to validate against Type, but we don't do it at
%% the moment.
default_value(T, S, _Type) ->
    {_Val, T1, S1} = scan_att_value(T, S).


%%%%%%% [71] EntityDef

scan_entity([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_entity(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_entity("%" ++ T, #xmerl_scanner{rules_write_fun = Write} = S0) ->
    %% parameter entity
    ?bump_col(1),
    ?strip1,
    {PEName, _NamespaceInfo, T2, S2} = scan_name_no_colons(T1, S1),
    ?strip3,
    {PEDef, T4, S4} = scan_pe_def(T3, S3, PEName),
    ?strip5,
    ">" ++ T6 = T5,
    S6 = Write(parameter_entity, PEName, PEDef, S5),
    {T6, S6};
scan_entity(T, #xmerl_scanner{rules_write_fun = Write} = S) ->
    %% generic entity
    {EName, _NamespaceInfo, T1, S1} = scan_name_no_colons(T, S),
    ?strip2,
    {EDef, T3, S3} = scan_entity_def(T2, S2, EName),
    ?strip4,
    ">" ++ T5 = T4,
    S5 = Write(entity, EName, EDef, S4),
    {T5, S5}.


%%%%%%% [73] EntityDef

scan_entity_def([], S = #xmerl_scanner{continuation_fun = F}, EName) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_entity_def(MoreBytes, S1, EName) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_entity_def("'" ++ T, S0, EName) ->
    ?bump_col(1),
    scan_entity_value(T, S, $', EName);
scan_entity_def("\"" ++ T, S0, EName) ->
    ?bump_col(1),
    scan_entity_value(T, S, $", EName);
scan_entity_def(Str, S, _EName) ->
    {ExtID, T1, S1} = scan_external_id(Str, S),
    ?strip2,
    {NData, T3, S3} = scan_ndata_decl(T2, S2),
    {{_USret,Entity},_Tail,_Sx}=fetch_and_parse(ExtID,S3,[{prolog,stop}]),
    {{Entity, NData}, T3, S3}.


scan_ndata_decl([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_ndata_decl(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_ndata_decl(Str = ">"++_T, S) ->
    {[], Str, S};
scan_ndata_decl("NDATA" ++ T, S0 = #xmerl_scanner{rules_read_fun = Read}) ->
    ?bump_col(5),
    ?strip1,
    {Name, _NamespaceInfo, T2, S2} = scan_name(T1, S1),
    case Read(notation, Name, S2) of
	undefined ->
	    ?fatal({unknown_notation, Name}, S2);
	_Value ->
	    {{ndata, Name}, T2, S2}
    end.

%%%%%%% [39] element

scan_element(T, S, Pos) ->
    scan_element(T, S, Pos, S#xmerl_scanner.space,
		 _Lang = [], _Parents = [], #xmlNamespace{}).

scan_element(T, S = #xmerl_scanner{line = L, col = C}, Pos, SpaceDefault, 
	     Lang, Parents, NS) ->
    {Name, NamespaceInfo, T1, S1} = scan_name(T, S),
    ?strip2,
    scan_element(T2, S2, Pos, Name, L, C, _Attrs = [], 
		 Lang, Parents, NamespaceInfo, NS, 
		 SpaceDefault).


scan_element("/", S = #xmerl_scanner{continuation_fun = F},
	     Pos, Name, StartL, StartC, Attrs, Lang, Parents, 
	     NSI, NS, SpaceDefault) ->
    ?dbg("trailing / detected~n", []),
    F(fun(MoreBytes, S1) -> scan_element("/" ++ MoreBytes, S1, 
					 Pos, Name, StartL, StartC, Attrs, 
					 Lang,Parents,NSI,NS,SpaceDefault) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_element([], S = #xmerl_scanner{continuation_fun = F}, 
	     Pos, Name, StartL, StartC, Attrs, Lang, Parents, 
	     NSI, NS, SpaceDefault) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_element(MoreBytes, S1, 
					 Pos, Name, StartL, StartC, Attrs, 
					 Lang,Parents,NSI,NS,SpaceDefault) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_element("/>" ++ T, S0 = #xmerl_scanner{hook_fun = Hook,
					    line = L, col = C,
					    event_fun = Event}, Pos,
	     Name, _StartL, _StartC, Attrs0, Lang, Parents, NSI, 
	     Namespace, _SpaceDefault) ->
    ?bump_col(2),
    Attrs = lists:reverse(Attrs0),
    E=processed_whole_element(S, Pos, Name, Attrs, Lang, Parents,NSI,Namespace),
    
    S1 = #xmerl_scanner{} = Event(#xmerl_event{event = ended,
					       line = L,
					       col = C,
					       data = E}, S0),
    {Ret, S2} = Hook(E, S1),
    {Ret, T, S2};
scan_element(">" ++ T, S0 = #xmerl_scanner{event_fun = Event,
					   line = L, col = C,
					   hook_fun = Hook,
					   space = SpaceOption},
	     Pos, Name, StartL, StartC, Attrs0, Lang, Parents, 
	     NSI, Namespace, SpaceDefault) ->
    ?bump_col(1),
    Attrs = lists:reverse(Attrs0),
    XMLSpace = case lists:keysearch('xml:space', 1, Attrs) of
		   false ->			SpaceDefault;
		   {value, "default"} ->	SpaceOption;
		   {value, "preserve"} ->	preserve;
		   _ ->				SpaceDefault
	       end,
    
    E0=processed_whole_element(S,Pos,Name,Attrs,Lang,Parents,NSI,Namespace),
    S1 = #xmerl_scanner{} = Event(#xmerl_event{event = started,
					       line = StartL,
					       col = StartC,
					       data = E0}, S),
    {Content, T1, S2} = scan_content(T, S1, Name, Attrs, XMLSpace, 
				     E0#xmlElement.language,
				     [{Name, Pos}|Parents], Namespace),
    Element=E0#xmlElement{content=Content,
			  directory=E0#xmlElement.directory},
    S3 = #xmerl_scanner{} = Event(#xmerl_event{event = ended,
					       line = L,
					       col = C,
					       data = Element}, S2),
    {Ret, S4} = Hook(Element, S3),
    {Ret, T1, S4};
scan_element(T, S, Pos, Name, StartL, StartC, Attrs, Lang, Parents, 
	     NSI, NS, SpaceDefault) ->
    {AttName, NamespaceInfo, T1, S1} = scan_name(T, S),
    {T2, S2} = scan_eq(T1, S1),
    {AttValue, T3, S3} = scan_att_value(T2, S2),
    NewNS = check_namespace(AttName, NamespaceInfo, AttValue, NS),
    ?strip4,
    AttrPos = case Attrs of
		  [] ->
		      1;
		  [#xmlAttribute{pos = P}|_] ->
		      P+1
	      end,
    Attr = #xmlAttribute{name = AttName, 
			 pos = AttrPos,
			 language = Lang,
			 namespace = NamespaceInfo,
			 value = AttValue},
    XMLBase=if
		AttName=='xml:base' ->
		    resolve_relative_uri(AttValue,S3#xmerl_scanner.directory);
		true ->	
		    S3#xmerl_scanner.directory
	    end,

    #xmerl_scanner{event_fun = Event,
		   line = Line,
		   col = Col} = S4,
    S5 = Event(#xmerl_event{event = ended,
			    line = Line,
			    col = Col,
			    data = Attr}, S4#xmerl_scanner{directory=XMLBase}),
    scan_element(T4, S5, Pos, Name, StartL, StartC, [Attr|Attrs], 
		 Lang, Parents, NSI, NewNS, SpaceDefault).


resolve_relative_uri(NewBase="/"++_,CurrentBase) ->
    case uri:parse(CurrentBase) of
	{error,no_scheme} ->
	    NewBase;
	{Scheme,Host,Port,PathQuery} ->
	    atom_to_list(Scheme)++Host++":"++integer_to_list(Port)++NewBase
    end;
resolve_relative_uri(NewBase,CurrentBase) ->
    NewBase.


processed_whole_element(S = #xmerl_scanner{hook_fun = _Hook,
					   directory = XMLBase,
					   line = _L, col = _C,
					   event_fun = _Event}, 
			Pos, Name, Attrs, Lang, Parents, NSI, Namespace) ->
    Language = check_language(Attrs, Lang),

    {ExpName, ExpAttrs} = 
	case S#xmerl_scanner.namespace_conformant of
	    true ->
		%% expand attribute names. We need to do this after having
		%% scanned all attributes of the element, since (as far as
		%% I can tell), XML Names only specifies that namespace attrs
		%% are valid within the whole scope of the element in which
		%% they are declared, which should also mean that even if they
		%% are declared after some other attributes, the namespace
		%% should apply to those attributes as well.
		%% Note that the default URI does not apply to attrbute names.
		TempNamespace = Namespace#xmlNamespace{default = []},
		ExpAttrsX = 
		    [A#xmlAttribute{
		       expanded_name=expanded_name(
				       A#xmlAttribute.name, 
				       A#xmlAttribute.namespace,
						% NSI,
				       TempNamespace, S)} || A <- Attrs],
		{expanded_name(Name, NSI, Namespace, S), ExpAttrsX};
	    false ->
		{Name, Attrs}
	end,

    #xmlElement{name = Name,
		directory = XMLBase,
		pos = Pos,
		parents = Parents,
		attributes = ExpAttrs,
		language = Language,
		expanded_name = ExpName,
		nsinfo = NSI,
		namespace = Namespace}.


check_language([#xmlAttribute{name='xml:lang',value=Lang}|_], _) ->
    Lang;
check_language([_|T], Lang) ->
    check_language(T, Lang);
check_language([], Lang) ->
    Lang.


check_namespace(xmlns, _, Value, NS) ->
    NS#xmlNamespace{default = list_to_atom(Value)};
check_namespace(_, {"xmlns", Prefix}, Value, 
		NS = #xmlNamespace{nodes = Ns}) ->
    NS#xmlNamespace{nodes = keyreplaceadd(
			      Prefix, 1, Ns, {Prefix, list_to_atom(Value)})};
check_namespace(_, _, _, NS) ->
    NS.


expanded_name(Name, [], #xmlNamespace{default = []}, _S) ->
    Name;
expanded_name(Name, [], #xmlNamespace{default = URI}, _S) ->
    {URI, Name};
expanded_name(_Name, {"xmlns", Local}, _NS, _S) -> % CHECK THIS /JB
    {"xmlns",Local};
expanded_name(_Name, {Prefix, Local}, #xmlNamespace{nodes = Ns}, S) ->
    case lists:keysearch(Prefix, 1, Ns) of
	{value, {_, URI}} ->
	    {URI, list_to_atom(Local)};
	false ->
	    %% A namespace constraint of XML Names is that the prefix
	    %% must be declared
	    ?fatal({namespace_prefix_not_declared, Prefix}, S)
    end.
		    



keyreplaceadd(K, Pos, [H|T], Obj) when K == element(Pos, H) ->
    [Obj|T];
keyreplaceadd(K, Pos, [H|T], Obj) ->
    [H|keyreplaceadd(K, Pos, T, Obj)];
keyreplaceadd(_K, _Pos, [], Obj) ->
    [Obj].

%%%%%%% [10] AttValue

scan_att_value([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_att_value(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_att_value([H|T], S0) when H == $"; H == $' ->
    ?bump_col(1),
    {_Str, _T1, _S1} = scan_att_chars(T, S, H, []).

scan_att_chars([], S = #xmerl_scanner{continuation_fun = F}, H, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_att_chars(MoreBytes, S1, H, Acc) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_att_chars([H|T], S0, H, Acc) ->
    ?bump_col(1),
    {lists:reverse(Acc), T, S};
scan_att_chars("&" ++ T, S0, Delim, Acc) ->
    ?bump_col(1),
    {ExpRef, T1, S1} = scan_reference(T, S),
    scan_att_chars(ExpRef ++ T1, S1, Delim, Acc);
scan_att_chars("<" ++ T, S0, Delim, Acc) ->
    ?fatal(unexpected_char, S0);
scan_att_chars([H|T], S0, Delim, Acc) ->
    ?bump_col(1),
    scan_att_chars(T, S, Delim, [H|Acc]).

%%%%%%% [43] content

scan_content(T, S, Name, Attrs, Space, Lang, Parents, NS) ->
    scan_content(T, S, _Pos = 1, Name, Attrs, Space, 
                 Lang, Parents, NS, _Acc = []).

scan_content("<", S= #xmerl_scanner{continuation_fun = F},
            Pos, Name, Attrs, Space, Lang, Parents, NS, Acc) ->
    ?dbg("trailing < detected~n", []),
    F(fun(MoreBytes, S1) -> scan_content("<" ++ MoreBytes, S1, 
					 Pos, Name, Attrs, 
					 Space, Lang, Parents, NS, Acc) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_content([], S = #xmerl_scanner{continuation_fun = F}, 
             Pos, Name, Attrs, Space, Lang, Parents, NS, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_content(MoreBytes, S1, 
					 Pos, Name, Attrs, 
					 Space, Lang, Parents, NS, Acc) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_content("</" ++ T, S0, _Pos, Name, _Attrs, _Space, _Lang,
	     _Parents, _NS, Acc) ->
    ?bump_col(2),
    {ETagName, _NamespaceInfo, T1, S1} = scan_name(T, S),
    if ETagName == Name ->
            ok;
       true ->
            ?fatal({endtag_does_not_match, {ETagName, Name}}, S1)
    end,
    ?strip2,
    case T2 of
        ">" ++ T3 ->
            {lists:reverse(Acc), T3, S2};
        _ -> % Mal-formed XML! Behave friendly and just skip to end of end tag.
            T3=tl(lists:dropwhile(fun(X) ->
                                          if X==$> -> false;
                                             true ->true
                                          end
                                  end,T2)),
            {lists:reverse(Acc), T3, S2}
    end;
scan_content("&" ++ T, S0, Pos, Name, Attrs, Space, Lang, Parents, NS, Acc) ->
    ?bump_col(1),
    {ExpRef, T1, S1} = scan_reference(T, S),
    scan_content(ExpRef++T1,S1,Pos,Name,Attrs,Space,Lang,Parents,NS,Acc);
scan_content("<!--" ++ T, S, Pos, Name, Attrs, Space, Lang, Parents, NS, Acc) ->
    {_, T1, S1} = scan_comment(T, S, Pos, Parents, Lang),
    scan_content(T1, S1, Pos+1, Name, Attrs, Space, Lang, Parents, NS, Acc);
scan_content("<" ++ T, S0, Pos, Name, Attrs, Space, Lang, Parents, NS, Acc) ->
    ?bump_col(1),
    {Markup, T1, S1} = 
        scan_content_markup(T, S, Pos, Name, Attrs, Space, Lang, Parents, NS),
    AccF = S1#xmerl_scanner.acc_fun,
    {NewAcc, NewPos, NewS} = case AccF(Markup, Acc, S1) of
				 {Acc2, S2} ->
				     {Acc2, Pos+1, S2};
				 {Acc2, Pos2, S2} ->
				     {Acc2, Pos2, S2}
			     end,
    scan_content(T1, S2, NewPos, Name, Attrs, Space, Lang,
		 Parents, NS, NewAcc);
scan_content(T, S = #xmerl_scanner{acc_fun = F, 
                                   event_fun = Event, 
                                   line = _L}, 
             Pos, Name, Attrs, Space, Lang, Parents, NS, Acc) ->
    Text0 = #xmlText{pos = Pos,
                     parents = Parents},
    S1 = #xmerl_scanner{} = Event(#xmerl_event{event = started,
                                               line = S#xmerl_scanner.line,
                                               data = Text0}, S),
    {Data, T1, S2} =  scan_char_data(T, S1, Space),
    Text = Text0#xmlText{value = Data},
    S3 = #xmerl_scanner{} = Event(#xmerl_event{event = ended,
                                               line = S2#xmerl_scanner.line,
                                               data = Text}, S2),
    {NewAcc, NewPos, NewS} = case F(Text, Acc, S3) of
				 {Acc4, S4} ->
				     {Acc4, Pos+1, S4};
				 {Acc4, Pos4, S4} ->
				     {Acc4, Pos4, S4}
			     end,
    scan_content(T1, NewS, NewPos, Name, Attrs, Space, Lang,
		 Parents, NS, NewAcc).


scan_content_markup([], S = #xmerl_scanner{continuation_fun = F}, 
		    Pos, Name, Attrs, Space, Lang, Parents, NS) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_content_markup(
			      MoreBytes,S1,Pos,Name,
			      Attrs,Space,Lang,Parents,NS) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_content_markup("![CDATA[" ++ T, S0, Pos, _Name, _Attrs,
		    _Space, _Lang, Parents, _NS) ->
    ?bump_col(8),
    scan_cdata(T, S, Pos, Parents);
scan_content_markup("?"++T,S0,Pos,_Name,_Attrs,_Space,_Lang,_Parents,_NS) ->
    ?bump_col(1),
    scan_pi(T, S, Pos);
scan_content_markup(T, S, Pos, _Name, _Attrs, Space, Lang, Parents, NS) ->
    scan_element(T, S, Pos, Space, Lang, Parents, NS).

scan_char_data(T, S, Space) ->
    scan_char_data(T, S, Space, _Acc = []).

%%%%%%% [14] CharData

scan_char_data([], S = #xmerl_scanner{continuation_fun = F}, Space, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_char_data(MoreBytes, S1, Space, Acc) end, 
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_char_data(T = [$&|_], S, _Space, Acc) ->
    {lists:reverse(Acc), T, S};
scan_char_data(T = [$<|_], S, _Space, Acc) ->
    {lists:reverse(Acc), T, S};
scan_char_data(T = [H|_], S, Space, Acc) when ?whitespace(H) ->
    {NewAcc, T1, S1} = accumulate_whitespace(T, S, Space, Acc),
    scan_char_data(T1, S1, Space, NewAcc);
scan_char_data([H|T], S0, Space, Acc) ->
    ?bump_col(1),
    scan_char_data(T, S, Space, [H|Acc]).



%%%%%%% [18]-[21] CDATA

scan_cdata(Str, S, Pos, Parents) ->
    scan_cdata(Str, S, Pos, Parents, _Acc = []).


scan_cdata([], S = #xmerl_scanner{continuation_fun = F}, Pos, Parents, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_cdata(MoreBytes, S1, Pos, Parents, Acc) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_cdata("]]>" ++ T, S0, Pos, Parents, Acc) ->
    ?bump_col(3),
    {#xmlText{pos = Pos,
	      parents = Parents,
	      value = lists:reverse(Acc)}, T, S};
scan_cdata([H|T], S0, Pos, Parents, Acc) ->
    ?bump_col(1),
    scan_cdata(T, S, Pos, Parents, [H|Acc]).


%%%%%%% [67] Reference


scan_reference([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_reference(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_reference("#x" ++ T, S0) ->
    %% [66] CharRef
    ?bump_col(1),
    if hd(T) /= $; ->
	    scan_char_ref_hex(T, S, 0);
       true ->
	    ?fatal(invalid_char_ref, S)
    end;
scan_reference("#" ++ T, S0) ->
    %% [66] CharRef
    ?bump_col(1),
    if hd(T) /= $; ->
	    scan_char_ref_dec(T, S, []);
       true ->
	    ?fatal(invalid_char_ref, S)
    end;
scan_reference(T, S) ->
    scan_entity_ref(T, S).


%% Chapter 4.4.2: ... the replacement text of entities used to escape
%% markup delimiters (the entities amp, lt, gt, apos, quot) is always treated
%% as data. (The string "AT&amp;T;" expands to "AT&T;" and the remaining
%% ampersand is not recognized as an entity-reference delimiter.)"
%%
%% How to achieve this? My current approach is to insert the *strings* "&",
%% "<", ">", "'", and "\"" instead of the characters. The processor will 
%% ignore them when performing multiple expansions. This means, for now, that
%% the character data output by the processor is (1-2 levels) deep.
%% At some suitable point, we should flatten these, so that application-level
%% processors should not have to be aware of this detail.

scan_entity_ref([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_entity_ref(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_entity_ref("amp;" ++ T, S0) ->
    ?bump_col(4),
    {["&"], T, S};
scan_entity_ref("lt;" ++ T, S0) ->
    ?bump_col(3),
    {["<"], T, S};
scan_entity_ref("gt;" ++ T, S0) ->
    ?bump_col(3),
    {[">"], T, S};
scan_entity_ref("apos;" ++ T, S0) ->
    ?bump_col(5),
    {["'"], T, S};
scan_entity_ref("quot;" ++ T, S0) ->
    ?bump_col(5),
    {["\""], T, S};
scan_entity_ref(T, S) ->
    {Name, _NamespaceInfo, T1, S1} = scan_name(T, S),
    ";" ++ T2 = T1, 
    S2 = S1,
    {Entity,_NData}=expand_reference(Name, S2),
    {Entity, T2, S2}.


%%%%%%% [69] PEReference

scan_pe_reference(T, S) ->
    {Name, _NamespaceInfo, T1, S1} = scan_name(T, S),
    ";" ++ T2 = T1, 
    {Name, T2, S1#xmerl_scanner{col = S1#xmerl_scanner.col+1}}.

expand_pe_reference(Name, #xmerl_scanner{rules_read_fun = Read} = S) ->
    case Read(parameter_entity, Name, S) of
	undefined ->
	    ?fatal({unknown_parameter_entity, Name}, S);
	Result ->
	    Result
    end.

expand_external_pe_reference(Name, #xmerl_scanner{rules_read_fun = Read} = S) ->
    case Read(parameter_entity, Name, S) of
	undefined ->
	    ?fatal({unknown_parameter_entity, Name}, S);
	Result ->
	    fetch_DTD(Result,S)
    end.
    

%%%%%%% [68] EntityReference

expand_reference(Name, #xmerl_scanner{rules_read_fun = Read} = S) ->
    case Read(entity, Name, S) of
	undefined ->
	    ?fatal({unknown_entity_ref, Name}, S);
	Value ->
	    Value
    end.


%%%%%%% [66] CharRef

scan_char_ref_dec([], S = #xmerl_scanner{continuation_fun = F}, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_char_ref_dec(MoreBytes, S1, Acc) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_char_ref_dec([H|T], S0, Acc) when H >= $0, H =< $9 ->
    ?bump_col(1),
    scan_char_ref_dec(T, S, [H|Acc]);
scan_char_ref_dec(";" ++ T, S0, Acc) ->
    ?bump_col(1),
    Ref = list_to_integer(lists:reverse(Acc)),
    {[[Ref]], T, S}.


scan_char_ref_hex([], S = #xmerl_scanner{continuation_fun = F}, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_char_ref_hex(MoreBytes, S1, Acc) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_char_ref_hex([H|T], S0, Acc) when H >= $0, H =< $9 ->
    ?bump_col(1),
    Dec = H - $0,
    scan_char_ref_hex(T, S, (Dec bor (Acc bsl 4)));
scan_char_ref_hex([H|T], S0, Acc) when H >= $a, H =< $f ->
    ?bump_col(1),
    Dec = (H - $a) + 10,
    scan_char_ref_hex(T, S, (Dec bor (Acc bsl 4)));
scan_char_ref_hex([H|T], S0, Acc) when H >= $A, H =< $F ->
    ?bump_col(1),
    Dec = (H - $A) + 10,
    scan_char_ref_hex(T, S, (Dec bor (Acc bsl 4)));
scan_char_ref_hex(";" ++ T, S0, Acc) ->
    ?bump_col(1),
    {[[Acc]], T, S}.



%%%%%%% [25] Eq
%%% Eq    ::=    S? '=' S?
scan_eq(T, S) ->
    ?strip1,
    "=" ++ T2 = T1, 
    S2 = S1,
    ?strip3,
    {T3, S3}.


%% scan_name/2
%%
%% We perform some checks here to make sure that the names conform to 
%% the "Namespaces in XML" specification. This is an option.
%% 
%% Qualified Name:
%% [6]      QName ::= (Prefix ':')? LocalPart
%% [7]     Prefix ::= NCName
%% [8]  LocalPart ::= NCName
%% [4]     NCName ::= (Letter | '_') (NCNameChar)*
%% [5] NCNameChar ::= Letter | Digit | '.' | '-' | '_' 
%%                    | CombiningChar | Extender


%% The effect of XML Names (namespace) conformance is that:
%% - All element types and attribute names contain either zero or one colon
%% - No entity names, PI targets, or notation names contain any colons.
%%
%% scan_name_no_colons/2 will ensure that the name contains no colons iff
%% the scanner has been told to be namespace conformant. Otherwise, it will
%% behave exactly like scan_name/2.
%%
scan_name_no_colons(Str, S) ->
    NSC = S#xmerl_scanner.namespace_conformant,
    case NSC of 
	true ->
	    {Target, NSI, T1, S1} = 
		scan_name(Str,S#xmerl_scanner{namespace_conformant=no_colons}),
	    {Target,NSI,T1,S1#xmerl_scanner{namespace_conformant=NSC}};
	false ->
	    scan_name(Str, S)
    end.



%% [5] Name ::= (Letter | '_' | ':') (NameChar)*
scan_name([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_name(MoreBytes, S1) end, 
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_name(Str = [$:|T], S0 = #xmerl_scanner{namespace_conformant = NSC}) ->
    if NSC == false ->
	    ?bump_col(1),
	    scan_nmtoken(T, S, [$:], NSC);
       NSC == no_colons ->
	    ?fatal({invalid_NCName, lists:sublist(Str, 1, 6)}, S0);
       true ->
	    %% In order to conform with the "Namespaces in XML" spec,
	    %% we cannot allow names to begin with ":"
	    ?fatal({invalid_NCName, lists:sublist(Str, 1, 6)}, S0)
    end;
scan_name([$_|T], S0 = #xmerl_scanner{namespace_conformant = NSC}) ->
    ?bump_col(1),
    scan_nmtoken(T, S, [$_], NSC);
scan_name([H|T], S0 = #xmerl_scanner{namespace_conformant = NSC}) ->
    case xmerl_lib:is_letter(H) of
	true ->
	    ?bump_col(1),
	    scan_nmtoken(T, S, [H], NSC);
	false ->
	    ?fatal({invalid_name, lists:sublist([H|T], 1, 6)}, S0)
    end;
scan_name(Str, S) ->
    ?fatal({invalid_name, Str}, S).






scan_nmtoken(Str, S, Acc, NSC) ->
    scan_nmtoken(Str, S, Acc, _Prefix = [], _Local = Acc, NSC).

%% scan_nmtoken/2
%% [7] NmToken ::= (NameChar)+
scan_nmtoken([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_nmtoken(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_nmtoken(Str = [H|T], S) ->
    case xmerl_lib:is_namechar(H) of
	true ->
	    scan_nmtoken(T, S#xmerl_scanner{col = S#xmerl_scanner.col+1}, 
			 _Acc = [H], _Prefix = [], _Local = [H], 
			 _NamespaceConformant = false);
	false ->
	    ?fatal({invalid_nmtoken, lists:sublist(Str, 1, 6)}, S)
    end.


scan_nmtoken([], S = #xmerl_scanner{continuation_fun = F}, 
	     Acc, Prefix, Local, NSC) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_nmtoken(MoreBytes,S1,Acc,Prefix,Local,NSC) end,
      fun(S1) -> {list_to_atom(lists:reverse(Acc)),
		  namespace_info(Prefix, Local),[],S1} end,
      S);
%% whitespace marks the end of a name
scan_nmtoken(Str = [H|_], S, Acc, Prefix, Local, _NSC) when ?whitespace(H) ->
    %% we don't strip here because the occurrence of whitespace may be an error
    %% e.g. <!ELEMENT spec (front, body, back ?)>
    NmString = lists:reverse(Acc),
    {list_to_atom(NmString), namespace_info(Prefix, Local), Str, S};

scan_nmtoken(Str = [$:|_], S, Acc, [], _Local, no_colons) ->
    ?fatal({invalid_NCName, 
	    lists:sublist(lists:reverse(Acc) ++ Str, 1, 6)}, S);
scan_nmtoken([$:|T], S0, Acc, [], Local, NSC) ->
    ?bump_col(1),
    scan_nmtoken(T, S, [$:|Acc], lists:reverse(Local), [], NSC);
scan_nmtoken(Str = [$:|_T], S, Acc, _Prefix, _Local, _NSC = true) ->
    %% non-empty Prefix means that we've encountered a ":" already.
    %% Conformity with "Namespaces in XML" requires 
    %% at most one colon in a name
    ?fatal({invalid_NCName, 
	    lists:sublist(lists:reverse(Acc) ++ Str, 1, 6)}, S);

%% non-namechar also marks the end of a name
scan_nmtoken(Str = [H|T], S0, Acc, Prefix, Local, NSC) ->
    ?bump_col(1),
    case xmerl_lib:is_namechar(H) of
	true ->
	    scan_nmtoken(T, S, [H|Acc], Prefix, [H|Local], NSC);
	false ->
	    NmStr = lists:reverse(Acc),
	    {list_to_atom(NmStr), namespace_info(Prefix, Local), Str, S}
    end.

namespace_info([], _) ->
    [];
namespace_info(Prefix, Local) ->
    {Prefix, lists:reverse(Local)}.

%%%%%%% [11] SystemLiteral

scan_system_literal([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_system_literal(MoreBytes, S1) end, 
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_system_literal("\"" ++ T, S) ->
    scan_system_literal(T, S, $", []);
scan_system_literal("'" ++ T, S) ->
    scan_system_literal(T, S, $', []).


scan_system_literal([], S = #xmerl_scanner{continuation_fun = F}, 
		    Delimiter, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_system_literal(MoreBytes,S1,Delimiter,Acc) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_system_literal([H|T], S, H, Acc) ->
    {lists:reverse(Acc), T, S#xmerl_scanner{col = S#xmerl_scanner.col+1}};
scan_system_literal([H|T], S, Delimiter, Acc) ->
    scan_system_literal(T, S#xmerl_scanner{col = S#xmerl_scanner.col+1}, 
			Delimiter, [H|Acc]).


%%%%%%% [12] PubidLiteral

scan_pubid_literal([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_pubid_literal(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_pubid_literal([H|T], S) when H == $"; H == $' ->
    scan_pubid_literal(T, S#xmerl_scanner{col = S#xmerl_scanner.col+1}, H, []);
scan_pubid_literal([H|_T], S) ->
    ?fatal({invalid_pubid_char, H}, S).


scan_pubid_literal([], S = #xmerl_scanner{continuation_fun = F}, 
		   Delimiter, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_pubid_literal(MoreBytes,S1,Delimiter,Acc) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_pubid_literal([H|T], S, H, Acc) ->
    {lists:reverse(Acc), T, S#xmerl_scanner{col = S#xmerl_scanner.col+1}};
scan_pubid_literal(Str = [H|_], S, Delimiter, Acc) when ?whitespace(H) ->
    %% Before matching public identifiers, all whitespace must be normalized,
    %% so we do that here
    {_, T, S1} = strip(Str, S),
    scan_pubid_literal(T, S1, Delimiter, [16#20|Acc]);
scan_pubid_literal([H|T], S, Delimiter, Acc) ->
    case is_pubid_char(H) of
	true ->
	    scan_pubid_literal(
	      T, S#xmerl_scanner{col = S#xmerl_scanner.col+1}, 
	      Delimiter, [H|Acc]);
	false ->
	    ?fatal({invalid_pubid_char, H}, S)
    end.

%% We do not match whitespace here, even though they're allowed in public
%% identifiers. This is because we normalize this whitespace as we scan
%% (see above in scan_pubid_literal())
%%
is_pubid_char(X) when X >= $a, X =< $z -> true;
is_pubid_char(X) when X >= $A, X =< $Z -> true;
is_pubid_char(X) when X >= $0, X =< $9 -> true;
is_pubid_char(X) ->
    lists:member(X, "-'()+,./:=?;!*#@$_%").


%%%%%%% [46] contentspec

scan_contentspec([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_contentspec(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_contentspec("EMPTY" ++ T, S0) ->
    ?bump_col(5),
    {empty, T, S};
scan_contentspec("ANY" ++ T, S0) ->
    ?bump_col(3),
    {any, T, S};
scan_contentspec("%" ++ T, S0) ->
    ?bump_col(1),
    {PERefName, T1, S1} = scan_pe_reference(T, S),
    ExpRef = expand_pe_reference(PERefName, S1),
    scan_contentspec(ExpRef ++ T1, S1);
scan_contentspec("(" ++ T, S0) ->
    ?bump_col(1),
    ?strip1,
    scan_elem_content(T1, S1).


%%%%%%% [47] children
%%%%%%% [51] Mixed

scan_elem_content(T, S) ->
    scan_elem_content(T, S, _Context = children, _Mode = unknown, _Acc = []).

scan_elem_content([], S = #xmerl_scanner{continuation_fun = F}, 
		  Context, Mode, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes,S1) -> scan_elem_content(MoreBytes,S1,Context,Mode,Acc) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_elem_content(")" ++ T, S0, Context, Mode0, Acc0) ->
    ?bump_col(1),
    {Mode, Acc} = case {Mode0, Acc0} of
		      {unknown, [_X]} ->
			  {seq, Acc0};
		      {M, _L} when M == seq; M == choice ->
			  {Mode0, lists:reverse(Acc0)}
		  end,
    {Occurrence, T1, S1} = scan_occurrence(T, S),
    case {Occurrence, Context} of
	{once, mixed} -> ok;
	{'*', mixed} -> ok;
	{Other, mixed} -> 
	    ?fatal({illegal_for_mixed_content, Other}, S1);
	_ ->
	    ok
    end,
    ?strip2,
    {format_elem_content({Occurrence, {Mode, Acc}}), T2, S2};
scan_elem_content("#PCDATA" ++ T, S0, _Context, Mode, Acc) ->
    ?bump_col(7),
    ?strip1,
    scan_elem_content(T1, S1, mixed, Mode, ['#PCDATA'|Acc]);
scan_elem_content("(" ++ T, S0, Context, Mode, Acc) ->
    ?bump_col(1),
    ?strip1,
    {Inner, T2, S2} = scan_elem_content(T1, S1),
    scan_elem_content(T2, S2, Context, Mode, [Inner|Acc]);
scan_elem_content("," ++ T, S0, Context, _Mode, Acc) ->
    ?bump_col(1),
    ?strip1,
    scan_elem_content(T1, S1, Context, seq, Acc);
scan_elem_content("|" ++ T, S0, Context, _Mode, Acc) ->
    ?bump_col(1),
    ?strip1,
    scan_elem_content(T1, S1, Context, choice, Acc);
scan_elem_content("%" ++ T, S0, Context, Mode, Acc) ->
    ?bump_col(1),
    {PERefName, T1, S1} = scan_pe_reference(T, S),
    ExpRef = expand_pe_reference(PERefName, S1),
    scan_elem_content(ExpRef ++ T1, S1, Context, Mode, Acc);
scan_elem_content(T, S, Context, Mode, Acc) ->
    {Name, _NameStr, T1, S1} = scan_name(T, S),
    {Occurrence, T2, S2} = scan_occurrence(T1, S1),
    case {Occurrence, Context} of
	{once, mixed} -> ok;
	{Other, mixed} -> 
	    ?fatal({illegal_for_mixed_content, Other}, S1);
	_ ->
	    ok
    end,
    ?strip3,
    NewAcc = [format_elem_content({Occurrence, Name}) | Acc],
    scan_elem_content(T3, S3, Context, Mode, NewAcc).

format_elem_content({once, What}) -> What;
format_elem_content(Other) -> Other.



scan_occurrence([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_occurrence(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_occurrence([$?|T], S0) ->
    ?bump_col(1),
    {'?', T, S};
scan_occurrence([$+|T], S0) ->
    ?bump_col(1),
    {'+', T, S};
scan_occurrence([$*|T], S0) ->
    ?bump_col(1),
    {'*', T, S};
scan_occurrence(T, S) ->
    {once, T , S}.


%%%%%%% [74] PEDef


scan_pe_def([], S = #xmerl_scanner{continuation_fun = F}, PEName) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_pe_def(MoreBytes, S1, PEName) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_pe_def("'" ++ T, S0, PEName) ->
    ?bump_col(1),
    scan_entity_value(T, S, $', PEName);
scan_pe_def("\"" ++ T, S0, PEName) ->
    ?bump_col(1),
    scan_entity_value(T, S, $", PEName);
scan_pe_def(Str, S, _PEName) ->
    scan_external_id(Str, S).


%%%%%%% [82] NotationDecl

scan_notation_decl(T, #xmerl_scanner{rules_write_fun = Write} = S) ->
    {Name, _NameStr, T1, S1} = scan_name_no_colons(T, S),
    ?strip2,
    {Def, T3, S3} = scan_notation_decl1(T2, S2),
    ?strip4,
    ">" ++ T5 = T4,
    S5 = Write(notation, Name, Def, S4),
    {T5, S5}.

scan_notation_decl1([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_notation_decl1(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_notation_decl1("SYSTEM" ++ T, S0) ->
    ?bump_col(6),
    ?strip1,
    {SL, T2, S2} = scan_system_literal(T1, S1),
    {{system, SL}, T2, S2};
scan_notation_decl1("PUBLIC" ++ T, S0) ->
    ?bump_col(6),
    ?strip1,
    {PIDL, T2, S2} = scan_pubid_literal(T1, S1),
    ?strip3,
    case T3 of
	">" ++ T4 ->
	    {{public, PIDL}, T4, 
	     S3#xmerl_scanner{col = S3#xmerl_scanner.col+1}};
	_ ->
	    {SL, T4, S4} = scan_system_literal(T3, S3),
	    {{public, PIDL, SL}, T4, S4}
    end.

%%%%%%% [75] ExternalID

scan_external_id([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_external_id(MoreBytes, S1) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_external_id("SYSTEM" ++ T, S0) ->
    ?bump_col(6),
    ?strip1,
    {SL, T2, S2} = scan_system_literal(T1, S1),
    {{system, SL}, T2, S2};
scan_external_id("PUBLIC" ++ T, S0) ->
    ?bump_col(6),
    ?strip1,
    {PIDL, T2, S2} = scan_pubid_literal(T1, S1),
    ?strip3,
    {SL, T4, S4} = scan_system_literal(T3, S3),
    {{public, PIDL, SL}, T4, S4}.


%%%%%%% [9] EntityValue

%% Note that we have two different scan functions for EntityValue
%% They differ in that this one checks for recursive calls to the same
%% parameter entity. This should perhaps be done

scan_entity_value(Str, S, Delim, Name) ->
    scan_entity_value(Str, S, Delim, _Acc = [], Name).


scan_entity_value([], S = #xmerl_scanner{continuation_fun = F}, 
		  H, Acc, PEName) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_entity_value(MoreBytes,S1,H,Acc,PEName) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_entity_value([H|T], S0, H, Acc, _PEName) ->
    ?bump_col(1),
    {lists:reverse(Acc), T, S};
scan_entity_value("%" ++ T, S0, Delim, Acc, PEName) ->
    ?bump_col(1),
    {PERefName, T1, S1} = scan_pe_reference(T, S),
    if PERefName == PEName ->
	    ?fatal({illegal_recursion_in_PE, PEName}, S1);
       true ->
	    ExpRef = expand_pe_reference(PERefName, S1),
	    scan_entity_value(ExpRef ++ T1, S1, Delim, Acc, PEName)
    end;
scan_entity_value("&" ++ T, S0, Delim, Acc, PEName) ->
    ?bump_col(1),
    {ExpRef, T1, S1} = scan_reference(T, S),
    scan_entity_value(ExpRef ++ T1, S1, Delim, Acc, PEName);
scan_entity_value([H|T], S0, Delim, Acc, PEName) ->
    ?bump_col(1),
    scan_entity_value(T, S, Delim, [H|Acc], PEName).


%%%%%%%



%%%%%%%



%%%%%%% [15] Comment
scan_comment(Str, S) ->
    scan_comment(Str, S, _Pos = undefined, _Parents = [], _Lang = []).

scan_comment(Str,S=#xmerl_scanner{col=C,event_fun=Event}, Pos, Parents, Lang) ->
    Comment = #xmlComment{pos = Pos,
			  parents = Parents,
			  language = Lang,
			  value = undefined},
    S1 = #xmerl_scanner{} = Event(#xmerl_event{event = started,
					       line = S#xmerl_scanner.line,
					       col = C,
					       pos = Pos,
					       data = Comment}, S),
    
    scan_comment1(Str, S1, Pos, Comment, _Acc = []).

scan_comment1([], S = #xmerl_scanner{continuation_fun = F}, 
	     Pos, Comment, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> scan_comment1(MoreBytes, S1, Pos, Comment, Acc) end,
      fun(S1) -> ?fatal(unexpected_end, S1) end,
      S);
scan_comment1("--" ++ T, S0 = #xmerl_scanner{col = C,
					     event_fun = Event,
					     hook_fun = Hook}, 
	     _Pos, Comment, Acc) ->
    case hd(T) of
	$> ->
	    ?bump_col(2),
	    Comment1 = Comment#xmlComment{value = lists:reverse(Acc)},
	    S1=#xmerl_scanner{}=Event(#xmerl_event{event = ended,
						   line=S#xmerl_scanner.line,
						   col = C,
						   data = Comment1}, S),
	    {Ret, S2} = Hook(Comment1, S1),
	    T2 = tl(T),
	    ?strip3,
	    {Ret, T3, S3};
	Char ->
	    ?fatal({invalid_comment,"--"++[Char]}, S0)
    end;
scan_comment1("\n" ++ T, S = #xmerl_scanner{line = L}, Pos, Cmt, Acc) ->
    scan_comment1(T, S#xmerl_scanner{line = L+1, col = 1}, 
		 Pos, Cmt, "\n" ++ Acc);
scan_comment1("\r\n" ++ T, S = #xmerl_scanner{line = L}, Pos, Cmt, Acc) ->
    %% CR followed by LF is read as a single LF
    scan_comment1(T, S#xmerl_scanner{line = L+1, col=1}, 
		  Pos, Cmt, "\n" ++ Acc);
scan_comment1("\r" ++ T, S = #xmerl_scanner{line = L}, Pos, Cmt, Acc) ->
    %% CR not followed by LF is read as a LF
    scan_comment1(T, S#xmerl_scanner{line = L+1, col = 1}, 
		 Pos, Cmt, "\n" ++ Acc);
scan_comment1([H|T], S = #xmerl_scanner{col = C}, Pos, Cmt, Acc) ->
    scan_comment1(T, S#xmerl_scanner{col = C+1}, Pos, Cmt, [H|Acc]).

%%%%%%%

strip([], S = #xmerl_scanner{continuation_fun = F}) ->
    ?dbg("cont()... stripping whitespace~n", []),
    F(fun(MoreBytes, S1) -> strip(MoreBytes, S1) end,
      fun(S1) -> {[], [], S1} end,
      S);
strip("\s" ++ T, S = #xmerl_scanner{col = C}) ->
    strip(T, S#xmerl_scanner{col = C+1});
strip("\t" ++ T, S = #xmerl_scanner{col = C}) ->
    strip(T, S#xmerl_scanner{col = expand_tab(C)});
strip("\n" ++ T, S = #xmerl_scanner{line = L}) ->
    strip(T, S#xmerl_scanner{line = L+1, col = 1});
strip("\r\n" ++ T, S = #xmerl_scanner{line = L}) ->
    %% CR followed by LF is read as a single LF
    strip(T, S#xmerl_scanner{line = L+1, col = 1});
strip("\r" ++ T, S = #xmerl_scanner{line = L}) ->
    %% CR not followed by LF is read as a LF
    strip(T, S#xmerl_scanner{line = L+1, col = 1});
strip(Str, S) ->
    {[], Str, S}.


%%% @spec accumulate_whitespace(T::string(),S::global_state(),
%%%                             atom(),Acc::string()) -> {Acc, T1, S1}
%%%     
%%% @doc Function to accumulate and normalize whitespace.
accumulate_whitespace(T, S, preserve, Acc) ->
    accumulate_whitespace(T, S, Acc);
accumulate_whitespace(T, S, normalize, Acc) ->
    {_WsAcc, T1, S1} = accumulate_whitespace(T, S, []),
    {[$\s|Acc], T1, S1}.


accumulate_whitespace([], S = #xmerl_scanner{continuation_fun = F}, Acc) ->
    ?dbg("cont()...~n", []),
    F(fun(MoreBytes, S1) -> accumulate_whitespace(MoreBytes, S1, Acc) end,
      fun(S1) -> {Acc, [], S1} end,
      S);
accumulate_whitespace("\s" ++ T, S = #xmerl_scanner{col = C}, Acc) ->
    accumulate_whitespace(T, S#xmerl_scanner{col = C+1}, [$\s|Acc]);
accumulate_whitespace("\t" ++ T, S = #xmerl_scanner{col = C}, Acc) ->
    accumulate_whitespace(T, S#xmerl_scanner{col = expand_tab(C)}, [$\t|Acc]);
accumulate_whitespace("\n" ++ T, S = #xmerl_scanner{line = L}, Acc) ->
    accumulate_whitespace(T, S#xmerl_scanner{line = L+1, col = 1}, [$\n|Acc]);
accumulate_whitespace("\r\n" ++ T, S = #xmerl_scanner{line = L}, Acc) ->
    %% CR followed by LF is read as a single LF
    accumulate_whitespace(T, S#xmerl_scanner{line = L+1, col=1}, [$\n|Acc]);
accumulate_whitespace("\r" ++ T, S = #xmerl_scanner{line = L}, Acc) ->
    %% CR not followed by LF is read as a LF
    accumulate_whitespace(T, S#xmerl_scanner{line = L+1, col = 1}, [$\n|Acc]);
accumulate_whitespace(Str, S, Acc) ->
    {Acc, Str, S}.


expand_tab(Col) ->
    Rem = (Col-1) rem 8,
    _NewCol = Col + 8 - Rem.

%%%%%%%% [2] Char

%is_char(16#9) -> true;
%is_char(16#a) -> true;
%is_char(16#d) -> true;
%is_char(X) when X >= 16#20, X =< 16#d7ff -> true;
%is_char(X) when X >= 16#e000, X =< 16#fffd -> true;
%is_char(X) when X >= 16#10000, X =< 16#10ffff -> true;
%is_char(_) ->
%    false.


%%%%%%%% [3] Whitespace

%is_whitespace(16#20) -> true;
%is_whitespace(16#9) -> true;
%is_whitespace(16#d) -> true;
%is_whitespace(16#a) -> true;
%is_whitespace(_) ->
%    false.

%%%%%%% [4] Namechar


%%% Helper functions

fatal(Reason, S) ->
    exit({fatal, {Reason, S#xmerl_scanner.line, S#xmerl_scanner.col}}).



%% BUG when we are many <!ATTLIST ..> balise none attributes has save in rules
rules_write(Context, Name, Value, #xmerl_scanner{rules = T} = S) ->
    case ets:lookup(T, {Context, Name}) of
	[] ->
	    ets:insert(T, {{Context, Name}, Value});
	_ ->
	    ok
    end,
    S.


rules_read(Context, Name, #xmerl_scanner{rules = T}) ->
    case ets:lookup(T, {Context, Name}) of
	[] ->
	    undefined;
	[{_, V}] ->
	    V
    end.

